@@ -47,6 +47,14 @@ struct _ShumateMapLayer
 
   ShumateMemoryCache *memcache;
 
+  gint64 profile_all_tiles_filled_begin;
+  gint64 profile_all_tiles_done_begin;
+
+  guint defer_callback_id;
+  double defer_latitude_y, defer_longitude_x, defer_zoom_level;
+  gint64 defer_frame_time;
+  gboolean deferring;
+
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
   ShumateVectorSymbolContainer *symbols;
 #endif
@@ -227,7 +235,7 @@ add_tile (ShumateMapLayer  *self,
       g_hash_table_insert (self->tile_fill, g_object_ref (tile), cancellable);
     }
 
-  g_hash_table_insert (self->tile_children, pos, g_object_ref (tile));
+  g_hash_table_insert (self->tile_children, pos, tile);
   gtk_widget_queue_draw (GTK_WIDGET (self));
   g_signal_connect_object (tile, "notify::state", (GCallback)on_tile_notify_state, self, G_CONNECT_SWAPPED);
 }
@@ -251,6 +259,68 @@ remove_tile (ShumateMapLayer  *self,
   g_signal_handlers_disconnect_by_func (tile, on_tile_notify_state, self);
 }
 
+static gboolean
+defer_tick_callback (GtkWidget     *widget,
+                     GdkFrameClock *frame_clock,
+                     gpointer       user_data)
+{
+  ShumateMapLayer *self = SHUMATE_MAP_LAYER (widget);
+  self->defer_callback_id = 0;
+
+  recompute_grid (self);
+  return G_SOURCE_REMOVE;
+}
+
+static gboolean
+should_defer (ShumateMapLayer *self)
+{
+  /* If the map is moving quickly, we may defer loading tiles until it slows back
+     down. That way, we don't waste resources loading tiles that will likely be gone
+     before they are done loading. */
+
+  ShumateViewport *viewport = shumate_layer_get_viewport (SHUMATE_LAYER (self));
+  ShumateMapSource *map_source = self->map_source;
+  double zoom_level = shumate_viewport_get_zoom_level (viewport);
+  double tile_size = shumate_map_source_get_tile_size_at_zoom (map_source, zoom_level);
+  double map_height = shumate_map_source_get_row_count (map_source, zoom_level) * tile_size;
+  double map_width = shumate_map_source_get_column_count (map_source, zoom_level) * tile_size;
+
+  double longitude_x = shumate_map_source_get_x (map_source, zoom_level, shumate_location_get_longitude (SHUMATE_LOCATION (viewport)));
+  double latitude_y = shumate_map_source_get_y (map_source, zoom_level, shumate_location_get_latitude (SHUMATE_LOCATION (viewport)));
+
+  double delta_x = self->defer_longitude_x * map_width - longitude_x;
+  double delta_y = self->defer_latitude_y * map_height - latitude_y;
+  double velocity = sqrt (delta_x * delta_x + delta_y * delta_y);
+  double width = gtk_widget_get_width (GTK_WIDGET (self));
+  double height = gtk_widget_get_height (GTK_WIDGET (self));
+  double diagonal = sqrt (width * width + height * height);
+  double zoom_velocity = self->defer_zoom_level - zoom_level;
+
+  gint64 frame_time = gdk_frame_clock_get_frame_time (gtk_widget_get_frame_clock (GTK_WIDGET (self)));
+
+  /* Only compare between frames, otherwise we might mistakenly think the
+     velocity is 0. */
+  if (frame_time == self->defer_frame_time)
+    return self->deferring;
+
+  if (velocity > diagonal * 0.25 || fabs (zoom_velocity) > 0.25)
+    {
+      if (self->defer_callback_id == 0)
+        self->defer_callback_id = gtk_widget_add_tick_callback (GTK_WIDGET (self), defer_tick_callback, NULL, NULL);
+
+      self->deferring = TRUE;
+    }
+  else
+    self->deferring = FALSE;
+
+  self->defer_latitude_y = latitude_y / map_height;
+  self->defer_longitude_x = longitude_x / map_width;
+  self->defer_zoom_level = shumate_viewport_get_zoom_level (viewport);
+  self->defer_frame_time = frame_time;
+
+  return self->deferring;
+}
+
 static void
 recompute_grid (ShumateMapLayer *self)
 {
@@ -273,6 +343,7 @@ recompute_grid (ShumateMapLayer *self)
   int source_columns = shumate_map_source_get_column_count (self->map_source, zoom_level);
 
   double rotation = shumate_viewport_get_rotation (viewport);
+  int n_tiles = 0;
 
   int size_x = MAX (
     abs ((int) (cos (rotation) *  width/2.0 - sin (rotation) * height/2.0)),
@@ -291,22 +362,26 @@ recompute_grid (ShumateMapLayer *self)
   int required_columns = tile_final_column - tile_initial_column;
   int required_rows = tile_final_row - tile_initial_row;
 
-  gboolean all_filled = TRUE;
+  gboolean defer = should_defer (self);
 
-  /* First, remove all the tiles that aren't in bounds. For now, ignore tiles
-   * that aren't on the current zoom level--those are only removed once the
-   * current level is fully loaded */
+  gboolean all_filled = TRUE, all_done = TRUE;
+
+  /* First, remove all the tiles that aren't in bounds, or that are on the
+   * wrong zoom level and haven't finished loading */
   g_hash_table_iter_init (&iter, self->tile_children);
   while (g_hash_table_iter_next (&iter, &key, &value))
     {
       TileGridPosition *pos = key;
       ShumateTile *tile = value;
+      float size = powf (2, zoom_level - pos->zoom);
+      float x = pos->x * size;
+      float y = pos->y * size;
 
-      if ((pos->x < tile_initial_column
-          || pos->x >= tile_initial_column + required_columns
-          || pos->y < tile_initial_row
-          || pos->y >= tile_initial_row + required_rows)
-          && pos->zoom == zoom_level)
+      if (x + size <= tile_initial_column
+          || x >= tile_initial_column + required_columns
+          || y + size <= tile_initial_row
+          || y >= tile_initial_row + required_rows
+          || (pos->zoom != zoom_level && shumate_tile_get_state (tile) != SHUMATE_STATE_DONE))
         {
           remove_tile (self, tile, pos);
           g_hash_table_iter_remove (&iter);
@@ -321,14 +396,19 @@ recompute_grid (ShumateMapLayer *self)
           g_autoptr(TileGridPosition) pos = tile_grid_position_new (x, y, zoom_level);
           ShumateTile *tile = g_hash_table_lookup (self->tile_children, pos);
 
-          if (!tile)
+          n_tiles ++;
+
+          if (!tile && !defer)
             {
               tile = shumate_tile_new_full (positive_mod (x, source_columns), positive_mod (y, source_rows), tile_size, zoom_level);
               add_tile (self, tile, g_steal_pointer (&pos));
             }
 
-          if (shumate_tile_get_state (tile) != SHUMATE_STATE_DONE)
+          if (tile == NULL || shumate_tile_get_paintable (tile) == NULL)
             all_filled = FALSE;
+
+          if (tile == NULL || shumate_tile_get_state (tile) != SHUMATE_STATE_DONE)
+            all_done = FALSE;
         }
     }
 
